"""
网格 Surface 类

用于定义接触对、载荷施加面等。

CalculiX 中 Surface 是节点集或单元面的集合，用于：
- 接触对定义（*CONTACT PAIR）
- 分布载荷（*DLOAD）
- 分布边界条件（*DSLOAD）
- 表面效应（*SFILM, *SFLUX）

类层次：
  SurfaceBase (ABC)
    ├── NodeSurface     # 节点基表面
    └── ElementSurface  # 单元面基表面

参考 pygccx mesh/surface.py 设计
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Iterable

from cae.enums import ElementType


# =============================================================================
# Surface 类型枚举
# =============================================================================

class SurfaceType(str):
    """Surface 类型。"""
    NODE = "NODE"       # 基于节点
    ELEMENT = "ELEMENT"  # 基于单元面


# =============================================================================
# Surface 基类
# =============================================================================

@dataclass
class SurfaceBase(ABC):
    """
    Surface 基类。

    Attributes:
        name: Surface 名称
    """

    name: str
    """Surface 名称"""

    @abstractmethod
    def to_inp_lines(self) -> list[str]:
        """转换为 INP 文件行。"""
        pass

    def __str__(self) -> str:
        return "\n".join(self.to_inp_lines())


# =============================================================================
# Node Surface
# =============================================================================

@dataclass
class NodeSurface(SurfaceBase):
    """
    基于节点的 Surface。

    适用于分布载荷、边界条件等。

    Attributes:
        name: Surface 名称
        node_ids: 节点 ID 集合
        node_set_names: 节点集名称集合
    """

    node_ids: set[int] = field(default_factory=set)
    """节点 ID 集合"""
    node_set_names: set[str] = field(default_factory=set)
    """节点集名称集合"""

    def to_inp_lines(self) -> list[str]:
        """转换为 INP 文件行。"""
        lines = [f"*SURFACE,NAME={self.name.upper()},TYPE={SurfaceType.NODE}"]

        # 先添加节点集名称
        for nset in self.node_set_names:
            lines.append(f"{nset},")

        # 再添加单独的节点 ID
        for nid in self.node_ids:
            lines.append(f"{nid},")

        # 删除最后一个逗号
        if lines[-1].endswith(","):
            lines[-1] = lines[-1][:-1]

        return lines

    def add_node_id(self, node_id: int) -> None:
        """添加节点 ID。"""
        self.node_ids.add(node_id)

    def add_node_set(self, nset_name: str) -> None:
        """添加节点集名称。"""
        self.node_set_names.add(nset_name)


# =============================================================================
# Element Surface
# =============================================================================

@dataclass
class ElementSurface(SurfaceBase):
    """
    基于单元面的 Surface。

    适用于接触对、表面载荷等。

    每个单元面由 (elem_id, face_id) 元组标识。
    face_id 是 CalculiX 手册中定义的面编号（从 1 开始）。

    Attributes:
        name: Surface 名称
        element_faces: 单元面集合 {(elem_id, face_id), ...}
    """

    element_faces: set[tuple[int, int]] = field(default_factory=set)
    """单元面集合 {(elem_id, face_id), ...}"""

    def to_inp_lines(self) -> list[str]:
        """转换为 INP 文件行。"""
        lines = [f"*SURFACE,NAME={self.name.upper()},TYPE={SurfaceType.ELEMENT}"]

        for elem_id, face_id in sorted(self.element_faces):
            lines.append(f"{elem_id},S{face_id}")

        return lines

    def add_element_face(self, elem_id: int, face_id: int) -> None:
        """添加单元面。"""
        self.element_faces.add((elem_id, face_id))


# =============================================================================
# Surface 工厂函数
# =============================================================================

def create_surface_from_node_set(
    name: str,
    elements: Iterable[ElementSurface],
    node_set_ids: set[int],
) -> ElementSurface:
    """
    从节点集创建单元面 Surface。

    查找所有包含给定节点集的单元及其面，构建 ElementSurface。

    Args:
        name: Surface 名称
        elements: 单元可迭代对象
        node_set_ids: 节点 ID 集合

    Returns:
        ElementSurface

    Note:
        需要 Element 类支持 get_faces() 方法。
        可以使用 cae.mesh.element 中的 Element 和 MeshElements。
    """
    from cae.mesh.element import get_element_faces

    surface: set[tuple[int, int]] = set()

    for elem in elements:
        elem_node_ids = set(elem.node_ids)

        # 检查节点集是否与单元有交集
        if node_set_ids.isdisjoint(elem_node_ids):
            continue

        # 检查每个面
        faces = get_element_faces(elem.type)
        if not faces:
            continue

        for face_idx, face_nodes in enumerate(faces):
            # face_idx 是 0-based，CalculiX face_id 是 1-based
            face_node_set = set(elem.node_ids[i] for i in face_nodes)
            if node_set_ids.issuperset(face_node_set):
                surface.add((elem.id, face_idx + 1))  # 转换为 1-based

    return ElementSurface(name=name.upper(), element_faces=surface)


def create_node_surface(
    name: str,
    node_ids: Iterable[int] = (),
    node_set_names: Iterable[str] = (),
) -> NodeSurface:
    """
    创建节点 Surface。

    Args:
        name: Surface 名称
        node_ids: 节点 ID 集合
        node_set_names: 节点集名称集合

    Returns:
        NodeSurface
    """
    return NodeSurface(
        name=name.upper(),
        node_ids=set(node_ids),
        node_set_names=set(node_set_names),
    )


def create_element_surface(
    name: str,
    element_faces: Iterable[tuple[int, int]] = (),
) -> ElementSurface:
    """
    创建单元面 Surface。

    Args:
        name: Surface 名称
        element_faces: 单元面集合 {(elem_id, face_id), ...}

    Returns:
        ElementSurface
    """
    return ElementSurface(
        name=name.upper(),
        element_faces=set(element_faces),
    )



"""
耦合 Coupling

定义运动耦合或分布耦合。

CalculiX 中 *COUPLING 与 *DISTRIBUTING 或 *KINEMATIC 配合使用，
将参考节点的运动/力耦合到单元面上。

参考 pygccx model_keywords/coupling.py 设计
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

from cae.enums import CouplingType
from cae.mesh.surface import ElementSurface


@dataclass
class Coupling:
    """
    耦合（运动或分布耦合）。

    将参考节点的运动或力耦合到单元面。

    Args:
        type: 耦合类型（DISTRIBUTING / KINEMATIC）
        ref_node: 参考节点 ID
        surface: 依赖表面（必须是单元面表面）
        name: 耦合名称
        first_dof: 第一个自由度（1-3 表示位移，4-6 表示旋转）
        last_dof: 可选，最后一个自由度（省略则仅使用 first_dof）
        orientation: 可选，局部坐标系名称
        cyclic_symmetry: 是否循环对称（仅 DISTRIBUTING 支持）
        desc: 描述文本

    Example:
        >>> from cae.coupling import Coupling
        >>> from cae.enums import CouplingType
        >>> from cae.mesh.surface import ElementSurface
        >>>
        >>> surf = ElementSurface(name='CoupSurf', element_faces={(1, 1), (2, 1)})
        >>> coup = Coupling(
        ...     type=CouplingType.KINEMATIC,
        ...     ref_node=100,
        ...     surface=surf,
        ...     name='KIN_COUP',
        ...     first_dof=1,
        ...     last_dof=3
        ... )
    """

    type: CouplingType
    """耦合类型"""
    ref_node: int
    """参考节点 ID"""
    surface: ElementSurface
    """依赖表面（必须是单元面表面）"""
    name: str
    """耦合名称"""
    first_dof: int
    """第一个自由度"""
    last_dof: Optional[int] = None
    """最后一个自由度"""
    orientation: Optional[str] = None
    """局部坐标系名称"""
    cyclic_symmetry: bool = False
    """是否循环对称（仅 DISTRIBUTING）"""
    desc: str = ""
    """描述文本"""
    keyword_name: str = "*COUPLING"
    """关键词名称"""

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "surface":
            if value.element_faces is None or len(value.element_faces) == 0:
                raise ValueError("surface 必须是单元面表面")
        if name == "first_dof":
            if value < 1 or value > 6:
                raise ValueError(f"first_dof 必须在 1-6 之间，当前值 {value}")
        if name == "last_dof" and value is not None:
            if value < 1 or value > 6:
                raise ValueError(f"last_dof 必须在 1-6 之间，当前值 {value}")
        super().__setattr__(name, value)

    def to_inp_lines(self) -> list[str]:
        """转换为 INP 文件行。"""
        lines = []

        # COUPLING 行
        line = f"*COUPLING,CONSTRAINT NAME={self.name},REF NODE={self.ref_node},SURFACE={self.surface.name}"
        if self.orientation:
            line += f",ORIENTATION={self.orientation}"
        if self.desc:
            line += f",DESCRIPTION={self.desc}"
        lines.append(line)

        # *DISTRIBUTING 或 *KINEMATIC
        line = self.type.value
        if self.type.value == "*DISTRIBUTING" and self.cyclic_symmetry:
            line += ",CYCLIC SYMMETRY"
        lines.append(line)

        # 自由度行
        dof_line = f"{self.first_dof}"
        if self.last_dof is not None:
            dof_line += f",{self.last_dof}"
        lines.append(dof_line)

        return lines

    def __str__(self) -> str:
        return "\n".join(self.to_inp_lines())



# 求解器注册表
"""
求解器注册表
管理所有可用求解器，提供实例化和查询接口。
"""
from __future__ import annotations

from typing import Optional

from .base import BaseSolver
from .calculix import CalculixSolver

# 求解器注册表
SOLVERS: dict[str, type[BaseSolver]] = {
    "calculix": CalculixSolver,
}


def get_solver(name: str) -> BaseSolver:
    """
    根据名称获取求解器实例。

    Args:
        name: 求解器名称（如 "calculix"）

    Returns:
        求解器实例

    Raises:
        ValueError: 如果求解器名称不存在
    """
    key = name.lower()
    if key not in SOLVERS:
        available = ", ".join(SOLVERS.keys())
        raise ValueError(f"未知求解器 '{name}'，可用求解器: {available}")
    return SOLVERS[key]()


def list_solvers() -> list[dict]:
    """
    列出所有已注册求解器信息。

    Returns:
        求解器信息列表，每项包含 name, installed, version, formats, description
    """
    result = []
    for name, cls in SOLVERS.items():
        solver = cls()
        result.append({
            "name": name,
            "installed": solver.check_installation(),
            "version": solver.get_version(),
            "formats": solver.supported_formats(),
            "description": solver.description,
        })
    return result



"""
MPC 多点约束

定义多点约束（Multiple Point Constraints）。

CalculiX 中 *MPC 用于定义节点之间的约束关系。

参考 pygccx model_keywords/mpc.py 设计
"""
from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from typing import Sequence

from cae.enums import MpcType


@dataclass
class Mpc:
    """
    MPC（多点约束）。

    定义节点之间的约束关系。

    Args:
        type: MPC 类型（PLANE / STRAIGHT / BEAM / MEANROT / DIST）
        nids: 参与的节点 ID 序列
        name: MPC 名称
        desc: 描述文本

    静态工厂方法：
        - Mpc.meanrot_from_node_set(node_set, n_pilot, name, desc)
        - Mpc.dist_from_3_nodes(n_a, n_b, n_dist, name, desc)
        - Mpc.beam_from_2_nodes(n_1, n_2, name, desc)
        - Mpc.straight_from_node_set(node_set, name, desc)
        - Mpc.plane_from_node_set(node_set, name, desc)

    Example:
        >>> from cae.coupling import Mpc
        >>> from cae.enums import MpcType
        >>>
        >>> # 刚体梁约束
        >>> mpc = Mpc(MpcType.BEAM, nids=[1, 2], name='rigid_beam')
        >>>
        >>> # 使用工厂方法
        >>> mpc = Mpc.beam_from_2_nodes(100, 101, name='pillar_to_slab')
    """

    type: MpcType
    """MPC 类型"""
    nids: Sequence[int]
    """参与的节点 ID 序列"""
    name: str = ""
    """MPC 名称"""
    desc: str = ""
    """描述文本"""
    keyword_name: str = "*MPC"
    """关键词名称"""

    def __post_init__(self):
        # 验证节点 ID
        for nid in self.nids:
            if nid < 1:
                raise ValueError(f"节点 ID 必须 >= 1，当前值 {nid}")

        # 类型特定验证
        if self.type == MpcType.BEAM:
            if len(self.nids) != 2:
                raise ValueError(f"BEAM 类型需要恰好 2 个节点，当前 {len(self.nids)} 个")

        elif self.type == MpcType.PLANE:
            if len(self.nids) < 3:
                raise ValueError(f"PLANE 类型需要至少 3 个节点，当前 {len(self.nids)} 个")

        elif self.type == MpcType.STRAIGHT:
            if len(self.nids) < 2:
                raise ValueError(f"STRAIGHT 类型需要至少 2 个节点，当前 {len(self.nids)} 个")

        elif self.type == MpcType.DIST:
            if len(self.nids) != 7:
                raise ValueError(f"DIST 类型需要恰好 7 个节点，当前 {len(self.nids)} 个")

        elif self.type == MpcType.MEANROT:
            if len(self.nids) < 4:
                raise ValueError(f"MEANROT 类型需要至少 4 个节点，当前 {len(self.nids)} 个")

        # MEANROT 和 DIST 的特殊验证
        if self.type in (MpcType.MEANROT, MpcType.DIST):
            # 直接验证每3个连续节点相同（这是最重要的验证）
            n_dep = len(self.nids) - 1  # 最后一个是pilot节点
            if n_dep % 3 != 0:
                raise ValueError(
                    f"{self.type.name} 类型：依赖节点数量 {n_dep} 不是 3 的倍数"
                )
            for i in range(n_dep // 3):
                start, end = 3 * i, 3 * (i + 1)
                if len(set(self.nids[start:end])) != 1:
                    raise ValueError(
                        f"{self.type.name} 类型：连续的 3 个节点必须是相同的，"
                        f"索引 {start} 到 {end} 不满足"
                    )
            # 最后一个节点（pilot）必须只出现 1 次
            if self.nids.count(self.nids[-1]) != 1:
                raise ValueError(
                    f"{self.type.name} 类型：最后一个节点（pilot）必须只出现 1 次"
                )

    def to_inp_lines(self) -> list[str]:
        """转换为 INP 文件行。"""
        lines = []

        if self.name:
            lines.append(f"*MPC,NAME={self.name}")
        else:
            lines.append("*MPC")
        if self.desc:
            lines.append(f"** {self.desc}")

        # 节点行：类型值 + 节点 ID
        nids = [self.type.value] + list(self.nids)

        # 每行最多 16 个值
        for i in range(0, len(nids), 16):
            temp = nids[i : i + 16]
            lines.append(",".join(map(str, temp)) + ",")

        # 删除最后一行末尾的逗号
        if lines[-1].endswith(","):
            lines[-1] = lines[-1][:-1]

        return lines

    def __str__(self) -> str:
        return "\n".join(self.to_inp_lines())

    @staticmethod
    def meanrot_from_node_set(
        node_ids: set[int], n_pilot: int, name: str = "", desc: str = ""
    ) -> "Mpc":
        """
        从节点集创建 MEANROT MPC。

        强制参考节点旋转为所有依赖节点旋转的平均值。

        Args:
            node_ids: 依赖节点 ID 集合
            n_pilot: 导向节点 ID
            name: MPC 名称
            desc: 描述文本

        Returns:
            Mpc: MEANROT MPC
        """
        # 每个节点重复3次，最后加导向节点
        nids = [n for nid in sorted(node_ids) for n in [nid] * 3] + [n_pilot]
        return Mpc(MpcType.MEANROT, nids, name, desc)

    @staticmethod
    def dist_from_3_nodes(
        n_a: int, n_b: int, n_dist: int, name: str = "", desc: str = ""
    ) -> "Mpc":
        """
        从 3 个节点创建 DIST MPC（最大距离约束）。

        强制节点 a 和 b 之间的距离不超过节点 dist 定义的值。

        Args:
            n_a: 定义方向的第一个节点
            n_b: 定义方向的第二个节点
            n_dist: 定义最大距离值的节点
            name: MPC 名称
            desc: 描述文本

        Returns:
            Mpc: DIST MPC
        """
        nids = [n_a] * 3 + [n_b] * 3 + [n_dist]
        return Mpc(MpcType.DIST, nids, name, desc)

    @staticmethod
    def beam_from_2_nodes(
        n_1: int, n_2: int, name: str = "", desc: str = ""
    ) -> "Mpc":
        """
        从 2 个节点创建刚体梁 MPC。

        强制两节点保持固定距离（刚性连接）。

        Args:
            n_1: 第一个节点
            n_2: 第二个节点
            name: MPC 名称
            desc: 描述文本

        Returns:
            Mpc: BEAM MPC
        """
        return Mpc(MpcType.BEAM, [n_1, n_2], name, desc)

    @staticmethod
    def straight_from_node_set(
        node_ids: set[int], name: str = "", desc: str = ""
    ) -> "Mpc":
        """
        从节点集创建 STRAIGHT MPC。

        强制所有依赖节点保持在一条直线上。

        Args:
            node_ids: 依赖节点 ID 集合
            name: MPC 名称
            desc: 描述文本

        Returns:
            Mpc: STRAIGHT MPC
        """
        nids = list(node_ids)
        return Mpc(MpcType.STRAIGHT, nids, name, desc)

    @staticmethod
    def plane_from_node_set(
        node_ids: set[int], name: str = "", desc: str = ""
    ) -> "Mpc":
        """
        从节点集创建 PLANE MPC。

        强制所有依赖节点保持在一个平面内。

        Args:
            node_ids: 依赖节点 ID 集合
            name: MPC 名称
            desc: 描述文本

        Returns:
            Mpc: PLANE MPC
        """
        nids = list(node_ids)
        return Mpc(MpcType.PLANE, nids, name, desc)



# cae package
"""
CAE-CLI: 轻量化 CAE 命令行工具

主要模块：
  - inp: INP 文件处理
  - solvers: 求解器接口
  - mesh: 网格处理
  - material: 材料模型
  - contact: 接触分析
  - coupling: 耦合约束
  - viewer: 可视化
  - ai: AI 助手

协议接口：
  - protocols: IKeyword, IStep 等接口定义
"""

from cae.protocols import IKeyword, IStep, INodeSet, IElementSet, ISurface



"""
Protocol 接口定义

参考 pygccx/protocols.py 设计，提供运行时检查的接口协议。

Protocol 特点：
- 使用 typing.Protocol 定义接口
- @runtime_checkable 装饰器支持 isinstance() 检查
- 子类只需实现协议方法，无需继承
"""
from __future__ import annotations

from typing import Protocol, runtime_checkable, Iterable


@runtime_checkable
class IKeyword(Protocol):
    """
    关键词协议。

    所有 INP 关键词类（如 Material, Elastic, Cload, Boundary）都应该实现此协议。

    实现要求：
      - keyword_name: str - 关键词名称（含 *，如 "*ELASTIC"）
      - desc: str - 描述文本（可选，用于注释）
      - to_inp_lines() -> list[str] - 转换为 INP 文件行

    Usage:
        def process_keyword(kw: IKeyword) -> None:
            if isinstance(kw, IKeyword):
                print(kw.to_inp_lines())

        # dataclass 自动满足协议
        @dataclass
        class Elastic:
            keyword_name: str = "*ELASTIC"
            params: dict = field(default_factory=dict)
            data: list = field(default_factory=list)

            def to_inp_lines(self) -> list[str]:
                ...

            def __str__(self) -> str:
                return "\\n".join(self.to_inp_lines())
    """

    @property
    def keyword_name(self) -> str:
        """关键词名称（含 *，如 "*ELASTIC", "*BOUNDARY"）。"""
        ...

    @property
    def desc(self) -> str:
        """描述文本，会作为注释写入 INP 文件。"""
        ...

    def to_inp_lines(self) -> list[str]:
        """
        转换为 INP 文件行列表。

        Returns:
            行列表，第一行是关键词行（如 "*ELASTIC,TYPE=ISO"），
            后续行是数据行。
        """
        ...


@runtime_checkable
class IStep(Protocol):
    """
    载荷步协议。

    表示一个分析步骤（如 *STEP ... *END STEP 之间的内容）。

    实现要求：
      - step_keywords: list[IKeyword] - 步内的关键词列表
      - add_step_keywords() - 添加步内关键词

    Usage:
        @dataclass
        class StaticStep:
            nlgeom: bool = False
            step_keywords: list[IKeyword] = field(default_factory=list)

            def add_step_keywords(self, *keywords: IKeyword) -> None:
                self.step_keywords.extend(keywords)
    """

    @property
    def step_keywords(self) -> list[IKeyword]:
        """步内的关键词列表。"""
        ...

    def add_step_keywords(self, *keywords: IKeyword) -> None:
        """
        添加关键词到步内。

        Args:
            *keywords: IKeyword 实例
        """
        ...


@runtime_checkable
class INodeSet(Protocol):
    """
    节点集协议。

    实现要求：
      - name: str - 集合名称
      - ids: set[int] - 节点 ID 集合
    """

    @property
    def name(self) -> str:
        """节点集名称。"""
        ...

    @property
    def ids(self) -> set[int]:
        """节点 ID 集合。"""
        ...


@runtime_checkable
class IElementSet(Protocol):
    """
    单元集协议。

    实现要求：
      - name: str - 集合名称
      - ids: set[int] - 单元 ID 集合
    """

    @property
    def name(self) -> str:
        """单元集名称。"""
        ...

    @property
    def ids(self) -> set[int]:
        """单元 ID 集合。"""
        ...


@runtime_checkable
class ISurface(Protocol):
    """
    表面协议。

    实现要求：
      - name: str - 表面名称
      - write_ccx(buffer: list[str]) - 写入 CCX 输入字符串到 buffer
    """

    @property
    def name(self) -> str:
        """表面名称。"""
        ...

    def write_ccx(self, buffer: list[str]) -> None:
        """
        将 CCX 输入字符串写入 buffer。

        Args:
            buffer: 字符串列表，会被就地修改
        """
        ...


# =============================================================================
# 类型别名
# =============================================================================

Number = int | float
"""数字类型别名。"""

KeywordIterable = Iterable[IKeyword]
"""关键词可迭代对象类型别名。"""



"""
绑定接触 Tie

定义绑定（tied）接触关系。

CalculiX 中 *TIE 用于绑定两个表面，使它们在分析过程中保持协调。

参考 pygccx model_keywords/tie.py 设计
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Optional, Union

from cae.mesh.surface import ElementSurface, NodeSurface
from cae._utils import f2s


@dataclass
class Tie:
    """
    绑定接触。

    将两个表面绑定在一起，使它们在分析过程中保持协调位移。

    Args:
        name: Tie 名称（最多 80 字符）
        dep_surf: 从表面（Dependent Surface）
        ind_surf: 主表面（Independent Surface）
        adjust: 是否调整绑定节点到主表面（默认 True）
        position_tolerance: 生成耦合的位置公差
        cyclic_symmetry: 是否用于循环对称模型
        multistage: 是否为多阶段耦合
        desc: 描述文本

    Example:
        >>> from cae.contact import Tie
        >>> from cae.mesh.surface import ElementSurface
        >>>
        >>> dep = ElementSurface(name="DEP_SURF", element_faces={(1, 1)})
        >>> ind = ElementSurface(name="IND_SURF", element_faces={(2, 2)})
        >>>
        >>> tie = Tie(
        ...     name="TIED_CONTACT",
        ...     dep_surf=dep,
        ...     ind_surf=ind,
        ...     adjust=True
        ... )
    """

    name: str
    """Tie 名称（最多 80 字符）"""
    dep_surf: Union[ElementSurface, NodeSurface]
    """从表面"""
    ind_surf: Union[ElementSurface, NodeSurface]
    """主表面"""
    adjust: bool = True
    """是否调整绑定节点到主表面"""
    position_tolerance: Optional[float] = None
    """生成耦合的位置公差"""
    cyclic_symmetry: bool = False
    """是否用于循环对称模型"""
    multistage: bool = False
    """是否为多阶段耦合"""
    desc: str = ""
    """描述文本"""
    keyword_name: str = "*TIE"
    """关键词名称"""

    _is_initialized: bool = field(init=False, default=False)

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        self._validate()

    def __post_init__(self):
        self._is_initialized = True

    def _validate(self) -> None:
        if not self._is_initialized:
            return

        if len(self.name) > 80:
            raise ValueError(f"name 最多 80 字符，当前 {len(self.name)} 字符")

        if self.position_tolerance is not None and self.position_tolerance < 0:
            raise ValueError(f"position_tolerance 必须 >= 0，当前值 {self.position_tolerance}")

        if self.cyclic_symmetry and self.multistage:
            raise ValueError("cyclic_symmetry 和 multistage 不能同时为 True")

        # 简单绑定接触：主表面必须是单元面
        if not self.cyclic_symmetry and not self.multistage:
            if self.ind_surf.element_faces is None or len(self.ind_surf.element_faces) == 0:
                raise ValueError("简单绑定接触时主表面(ind_surf)必须是单元面表面")

        # 多阶段耦合：两者都必须是节点表面
        if self.multistage:
            if self.ind_surf.node_ids is not None and len(self.ind_surf.node_ids) > 0:
                raise ValueError("多阶段耦合时主表面(ind_surf)必须是节点表面")
            if self.dep_surf.node_ids is not None and len(self.dep_surf.node_ids) > 0:
                raise ValueError("多阶段耦合时从表面(dep_surf)必须是节点表面")

    def to_inp_lines(self) -> list[str]:
        """转换为 INP 文件行。"""
        lines = []

        # TIE 行
        line = f"*TIE,NAME={self.name}"
        if not self.adjust:
            line += ",ADJUST=NO"
        if self.position_tolerance is not None:
            line += f",POSITION TOLERANCE={f2s(self.position_tolerance)}"
        if self.cyclic_symmetry:
            line += ",CYCLIC SYMMETRY"
        if self.multistage:
            line += ",MULTISTAGE"
        if self.desc:
            line += f",DESCRIPTION={self.desc}"
        lines.append(line)

        # 表面名称行
        lines.append(f"{self.dep_surf.name},{self.ind_surf.name}")

        return lines

    def __str__(self) -> str:
        return "\n".join(self.to_inp_lines())



"""
第二周：viewer 模块单元测试
覆盖：frd_parser、vtk_export、server（端口查找 / 文件收集）
所有 meshio I/O 均通过 mock 隔离，无需真实 .frd 文件。
"""
from __future__ import annotations

import json
from pathlib import Path
from unittest.mock import MagicMock, patch, mock_open

import numpy as np
import pytest

from cae.viewer.frd_parser import (
    FrdData,
    FrdElement,
    FrdNodes,
    FrdResultStep,
    parse_frd,
)
from cae.enums import FrdResultEntity
from cae.viewer.vtk_export import (
    VtkExportResult,
    frd_to_vtu,
)
from cae.viewer._utils import von_mises as _von_mises


# ================================================================== #
# frd_parser
# ================================================================== #

class TestFrdNodes:
    def test_basic_fields(self):
        n = FrdNodes(ids=[1, 2, 3], coords=[(0, 0, 0), (1, 0, 0), (0, 1, 0)])
        assert len(n.ids) == 3
        assert len(n.coords) == 3


class TestFrdData:
    def test_has_geometry_false_when_empty(self):
        d = FrdData()
        assert d.has_geometry is False

    def test_has_geometry_false_no_elements(self):
        d = FrdData(nodes=FrdNodes(ids=[1], coords=[(0, 0, 0)]))
        assert d.has_geometry is False

    def test_has_geometry_true(self):
        d = FrdData(
            nodes=FrdNodes(ids=[1, 2, 3, 4], coords=[(0,0,0)]*4),
            elements=[FrdElement(eid=1, etype=3, connectivity=[1,2,3,4])],
        )
        assert d.has_geometry is True

    def test_node_count(self):
        d = FrdData(nodes=FrdNodes(ids=[1, 2], coords=[(0,0,0),(1,0,0)]))
        assert d.node_count == 2

    def test_node_count_no_nodes(self):
        assert FrdData().node_count == 0

    def test_get_result_by_name(self):
        r1 = FrdResultStep(step=1, time=0.0, name="DISP", components=[], values={}, node_ids=[], entity=FrdResultEntity.DISP)
        r2 = FrdResultStep(step=1, time=0.0, name="STRESS", components=[], values={}, node_ids=[], entity=FrdResultEntity.STRESS)
        d = FrdData(results=[r1, r2])
        assert d.get_result("DISP") is r1
        assert d.get_result("STRESS") is r2

    def test_get_result_missing(self):
        d = FrdData()
        assert d.get_result("NONEXISTENT") is None

    def test_get_result_last_step(self):
        steps = [
            FrdResultStep(step=i, time=float(i), name="DISP", components=[], values={}, node_ids=[], entity=FrdResultEntity.DISP)
            for i in range(1, 4)
        ]
        d = FrdData(results=steps)
        assert d.get_result("DISP", step=-1) is steps[-1]
        assert d.get_result("DISP", step=0)  is steps[0]


class TestParseFrd:
    def test_file_not_found(self, tmp_path: Path):
        with pytest.raises(FileNotFoundError):
            parse_frd(tmp_path / "ghost.frd")

    def _write_minimal_frd(self, path: Path) -> Path:
        """最小化合法 ASCII .frd 文件（4节点四面体 + 位移结果）。"""
        content = """\
    1C                          4
 -1         1 0.000000E+00 0.000000E+00 0.000000E+00
 -1         2 1.000000E+00 0.000000E+00 0.000000E+00
 -1         3 0.000000E+00 1.000000E+00 0.000000E+00
 -1         4 0.000000E+00 0.000000E+00 1.000000E+00
 -3
    2C                          1
 -1         1  3  1  1
 -2         1  2  3  4
 -3
  100C                             DISP        1  0.00000E+00
 -4  DISP        4  0  1
 -5  D1          1  2  1  0
 -5  D2          1  2  2  0
 -5  D3          1  2  3  0
 -1         1 1.00000E-03 0.00000E+00 0.00000E+00
 -1         2 2.00000E-03 0.00000E+00 0.00000E+00
 -1         3 3.00000E-03 0.00000E+00 0.00000E+00
 -1         4 4.00000E-03 0.00000E+00 0.00000E+00
 -3
9999
"""
        path.write_text(content)
        return path

    def test_parse_nodes(self, tmp_path: Path):
        frd = self._write_minimal_frd(tmp_path / "test.frd")
        data = parse_frd(frd)
        assert data.node_count == 4

    def test_parse_elements(self, tmp_path: Path):
        frd = self._write_minimal_frd(tmp_path / "test.frd")
        data = parse_frd(frd)
        assert data.element_count == 1
        assert data.elements[0].etype == 3  # C3D4

    def test_parse_results(self, tmp_path: Path):
        frd = self._write_minimal_frd(tmp_path / "test.frd")
        data = parse_frd(frd)
        disp = data.get_result("DISP")
        assert disp is not None
        assert len(disp.node_ids) == 4

    def test_has_geometry(self, tmp_path: Path):
        frd = self._write_minimal_frd(tmp_path / "test.frd")
        data = parse_frd(frd)
        assert data.has_geometry is True


# ================================================================== #
# vtk_export
# ================================================================== #

class TestVonMises:
    def test_uniaxial_tension(self):
        """单轴拉伸：S11=100, 其他=0 → Von Mises = 100."""
        stress = np.array([[100.0, 0, 0, 0, 0, 0]])
        vm = _von_mises(stress)
        assert pytest.approx(vm[0], abs=1e-6) == 100.0

    def test_hydrostatic_zero(self):
        """静水压：S11=S22=S33=p, 剪切=0 → Von Mises = 0."""
        p = 50.0
        stress = np.array([[p, p, p, 0.0, 0.0, 0.0]])
        vm = _von_mises(stress)
        assert pytest.approx(vm[0], abs=1e-6) == 0.0

    def test_pure_shear(self):
        """纯剪切 S12=τ → Von Mises = √3 · τ."""
        tau = 100.0
        stress = np.array([[0.0, 0.0, 0.0, tau, 0.0, 0.0]])
        vm = _von_mises(stress)
        assert pytest.approx(vm[0], rel=1e-6) == tau * np.sqrt(3)

    def test_batch(self):
        """批量输入：输出长度与输入相同。"""
        stress = np.random.rand(50, 6)
        vm = _von_mises(stress)
        assert vm.shape == (50,)
        assert np.all(vm >= 0)


class TestVtkExportResult:
    def test_default_fields_list(self):
        r = VtkExportResult(success=True)
        assert isinstance(r.fields, list)

    def test_success_with_vtu(self, tmp_path: Path):
        vtu = tmp_path / "job.vtu"
        vtu.write_text("<VTKFile/>")
        r = VtkExportResult(success=True, vtu_file=vtu, node_count=8, element_count=1)
        assert r.success
        assert r.vtu_file == vtu


class TestFrdToVtu:
    def test_missing_frd_returns_error(self, tmp_path: Path):
        result = frd_to_vtu(tmp_path / "no.frd")
        assert result.success is False
        assert "不存在" in (result.error or "")

    def test_meshio_direct_success(self, tmp_path: Path):
        """meshio 直读成功路径。"""
        frd = tmp_path / "job.frd"
        frd.write_bytes(b"dummy")

        mock_mesh = MagicMock()
        mock_mesh.points = np.zeros((8, 3))
        mock_mesh.cells = [MagicMock(data=np.zeros((1, 8), dtype=int))]
        mock_mesh.point_data = {"U": np.zeros((8, 3))}
        mock_mesh.cell_data = {}

        mock_meshio = MagicMock()
        mock_meshio.read.return_value = mock_mesh

        with (
            patch("cae.viewer.vtk_export._meshio_module", mock_meshio),
            patch("cae.viewer.vtk_export._HAS_MESHIO", True),
        ):
            result = frd_to_vtu(frd, tmp_path)

        assert result.success is True
        assert result.node_count == 8

    def test_fallback_when_meshio_fails(self, tmp_path: Path):
        """meshio 直读失败时走内置解析器路径。"""
        frd = tmp_path / "job.frd"
        frd.write_text("""\
    1C                          2
 -1         1 0.000000E+00 0.000000E+00 0.000000E+00
 -1         2 1.000000E+00 0.000000E+00 0.000000E+00
 -3
    2C                          1
 -1         1  3  1  1
 -2         1  2  1  2
 -3
9999
""")
        mock_meshio = MagicMock()
        mock_meshio.read.side_effect = Exception("unsupported format")

        with (
            patch("cae.viewer.vtk_export._meshio_module", mock_meshio),
            patch("cae.viewer.vtk_export._HAS_MESHIO", True),
        ):
            result = frd_to_vtu(frd, tmp_path)

        assert isinstance(result, VtkExportResult)

    def test_output_dir_created(self, tmp_path: Path):
        """输出目录不存在时自动创建。"""
        frd = tmp_path / "job.frd"
        frd.write_bytes(b"dummy")
        out = tmp_path / "nested" / "output"

        mock_meshio = MagicMock()
        mock_meshio.read.side_effect = Exception("fail")

        with (
            patch("cae.viewer.vtk_export._meshio_module", mock_meshio),
            patch("cae.viewer.vtk_export._HAS_MESHIO", True),
        ):
            frd_to_vtu(frd, out)

        assert out.exists()


# ================================================================== #
# server utilities
# ================================================================== #

class TestServerUtils:
    def test_find_free_port_returns_int(self):
        from cae.viewer.server import _find_free_port
        port = _find_free_port(18888)
        assert isinstance(port, int)
        assert 18888 <= port < 18988

    def test_collect_vtu_files_empty(self, tmp_path: Path):
        from cae.viewer.server import _collect_vtu_files
        assert _collect_vtu_files(tmp_path) == []

    def test_collect_vtu_files_found(self, tmp_path: Path):
        from cae.viewer.server import _collect_vtu_files
        (tmp_path / "a.vtu").write_text("")
        (tmp_path / "b.vtk").write_text("")
        (tmp_path / "c.frd").write_text("")  # should not be included
        files = _collect_vtu_files(tmp_path)
        names = {f.name for f in files}
        assert "a.vtu" in names
        assert "b.vtk" in names
        assert "c.frd" not in names

    def test_start_server_dir_not_found(self, tmp_path: Path):
        from cae.viewer.server import start_server
        with pytest.raises(FileNotFoundError):
            start_server(tmp_path / "nonexistent", open_browser=False)

    def test_start_server_no_vtu_files(self, tmp_path: Path):
        from cae.viewer.server import start_server
        # 目录存在但没有可视化文件
        (tmp_path / "dummy.txt").write_text("")
        with pytest.raises(FileNotFoundError, match="没有可视化文件"):
            start_server(tmp_path, auto_convert=False, open_browser=False)

    def test_start_server_returns_server_and_url(self, tmp_path: Path):
        from cae.viewer.server import start_server
        (tmp_path / "result.vtu").write_text("<VTKFile/>")
        server, url, files = start_server(tmp_path, open_browser=False, auto_convert=False)
        assert url.startswith("http://localhost:")
        assert len(files) == 1
        server.server_close()


class TestIndexHtml:
    """验证 index.html 生成正确的文件按钮和 JSON。"""

    def test_file_buttons_in_html(self, tmp_path: Path):
        from cae.viewer.server import start_server
        (tmp_path / "beam.vtu").write_text("")
        (tmp_path / "frame.vtu").write_text("")

        server, url, files = start_server(tmp_path, open_browser=False, auto_convert=False)
        server.server_close()

        # 检查 handler 能获取正确的文件列表
        assert len(files) == 2
        names = {f.name for f in files}
        assert "beam.vtu" in names
        assert "frame.vtu" in names


# ================================================================== #
# 应力计算工具测试
# ================================================================== #

from cae.viewer._utils import (
    von_mises,
    get_principal_stresses,
    get_principal_shear_stresses,
    get_max_shear_stress,
    get_worst_principal_stress,
    get_stress_invariants,
)


class TestPrincipalStresses:
    """主应力计算测试"""

    def test_uniaxial_tension(self):
        """单轴拉伸：σ1=100, σ2=σ3=0"""
        stress = np.array([[100.0, 0, 0, 0, 0, 0]])
        principal, _ = get_principal_stresses(stress)
        assert pytest.approx(principal[0][0], abs=1e-6) == 100.0
        assert pytest.approx(principal[0][1], abs=1e-6) == 0.0
        assert pytest.approx(principal[0][2], abs=1e-6) == 0.0

    def test_uniaxial_compression(self):
        """单轴压缩：σ1=σ2=0, σ3=-100"""
        stress = np.array([[0, 0, -100.0, 0, 0, 0]])
        principal, _ = get_principal_stresses(stress)
        assert pytest.approx(principal[0][0], abs=1e-6) == 0.0
        assert pytest.approx(principal[0][1], abs=1e-6) == 0.0
        assert pytest.approx(principal[0][2], abs=1e-6) == -100.0

    def test_pure_shear(self):
        """纯剪切：σ1=τ, σ2=0, σ3=-τ"""
        tau = 50.0
        stress = np.array([[0, 0, 0, tau, 0, 0]])
        principal, _ = get_principal_stresses(stress)
        assert pytest.approx(principal[0][0], abs=1e-6) == tau
        assert pytest.approx(principal[0][1], abs=1e-6) == 0.0
        assert pytest.approx(principal[0][2], abs=1e-6) == -tau

    def test_hydrostatic_pressure(self):
        """静水压：σ1=σ2=σ3=-p"""
        p = 100.0
        stress = np.array([[-p, -p, -p, 0, 0, 0]])
        principal, _ = get_principal_stresses(stress)
        assert all(abs(principal[0][i] - (-p)) < 1e-10 for i in range(3))

    def test_principal_directions(self):
        """验证主方向是正交的"""
        stress = np.array([[100.0, 50.0, 0, 10.0, 0, 0]])
        _, directions = get_principal_stresses(stress)
        d = directions[0]  # (3, 3)
        # 验证每列是单位向量
        for col in range(3):
            norm = np.linalg.norm(d[:, col])
            assert pytest.approx(norm, abs=1e-6) == 1.0
        # 验证列之间正交
        dot01 = np.dot(d[:, 0], d[:, 1])
        dot02 = np.dot(d[:, 0], d[:, 2])
        dot12 = np.dot(d[:, 1], d[:, 2])
        assert pytest.approx(dot01, abs=1e-6) == 0.0
        assert pytest.approx(dot02, abs=1e-6) == 0.0
        assert pytest.approx(dot12, abs=1e-6) == 0.0

    def test_batch(self):
        """批量计算：结果形状正确"""
        stress = np.random.rand(20, 6)
        principal, directions = get_principal_stresses(stress)
        assert principal.shape == (20, 3)
        assert directions.shape == (20, 3, 3)


class TestPrincipalShearStresses:
    """主剪切应力计算测试"""

    def test_pure_shear(self):
        """纯剪切 S12=τ → τ12=τ/2, τ13=τ, τ23=τ/2"""
        tau = 100.0
        stress = np.array([[0, 0, 0, tau, 0, 0]])
        shear = get_principal_shear_stresses(stress)
        assert pytest.approx(shear[0][0], abs=1e-6) == tau / 2  # |σ1-σ2|/2 = |τ|/2
        # 对于纯剪切，主应力是 [τ, 0, -τ]
        # τ12 = |τ - 0|/2 = τ/2
        # τ13 = |τ - (-τ)|/2 = τ
        # τ23 = |0 - (-τ)|/2 = τ/2

    def test_uniaxial_tension(self):
        """单轴拉伸：σ1=σ, σ2=σ3=0"""
        sigma = 100.0
        stress = np.array([[sigma, 0, 0, 0, 0, 0]])
        shear = get_principal_shear_stresses(stress)
        assert pytest.approx(shear[0][0], abs=1e-6) == sigma / 2  # |σ1-σ2|/2
        assert pytest.approx(shear[0][1], abs=1e-6) == sigma / 2  # |σ1-σ3|/2
        assert pytest.approx(shear[0][2], abs=1e-6) == 0.0  # |σ2-σ3|/2

    def test_batch(self):
        """批量计算"""
        stress = np.random.rand(30, 6)
        shear = get_principal_shear_stresses(stress)
        assert shear.shape == (30, 3)
        assert np.all(shear >= 0)


class TestMaxShearStress:
    """最大剪切应力测试"""

    def test_pure_shear(self):
        """纯剪切：最大剪切应力等于最大主应力"""
        tau = 50.0
        stress = np.array([[0, 0, 0, tau, 0, 0]])
        max_shear = get_max_shear_stress(stress)
        assert pytest.approx(max_shear[0], abs=1e-6) == tau

    def test_uniaxial_tension(self):
        """单轴拉伸：最大剪切应力 = σ1/2"""
        sigma = 200.0
        stress = np.array([[sigma, 0, 0, 0, 0, 0]])
        max_shear = get_max_shear_stress(stress)
        assert pytest.approx(max_shear[0], abs=1e-6) == sigma / 2

    def test_batch(self):
        """批量计算"""
        stress = np.random.rand(25, 6)
        max_shear = get_max_shear_stress(stress)
        assert max_shear.shape == (25,)
        assert np.all(max_shear >= 0)


class TestWorstPrincipalStress:
    """最不利主应力测试"""

    def test_tension_worst_is_positive(self):
        """受拉时最不利是最大主应力（正值）"""
        stress = np.array([[100.0, 0, 0, 0, 0, 0]])
        worst = get_worst_principal_stress(stress)
        assert pytest.approx(worst[0], abs=1e-6) == 100.0

    def test_compression_worst_is_negative(self):
        """受压时最不利是最小主应力（负值）"""
        stress = np.array([[0, 0, -100.0, 0, 0, 0]])
        worst = get_worst_principal_stress(stress)
        assert pytest.approx(worst[0], abs=1e-6) == -100.0

    def test_mixed(self):
        """拉压混合时比较绝对值"""
        # σ1=100 (tension), σ3=-50 (compression) → 最不利是 100
        stress = np.array([[50.0, 0, -30.0, 0, 0, 0]])
        worst = get_worst_principal_stress(stress)
        assert pytest.approx(worst[0], abs=1e-6) == 50.0


class TestStressInvariants:
    """应力不变量测试"""

    def test_hydrostatic_pressure(self):
        """静水压 I1 = -3p, I2 = 3p², I3 = -p³"""
        p = 50.0
        stress = np.array([[-p, -p, -p, 0, 0, 0]])
        inv = get_stress_invariants(stress)
        assert pytest.approx(inv["I1"][0], abs=1e-6) == -3 * p
        assert pytest.approx(inv["I2"][0], abs=1e-6) == 3 * p ** 2
        assert pytest.approx(inv["I3"][0], abs=1e-6) == -p ** 3

    def test_uniaxial_tension(self):
        """单轴拉伸：σ1=σ, I1=σ, I2=-σ², I3=0"""
        sigma = 100.0
        stress = np.array([[sigma, 0, 0, 0, 0, 0]])
        inv = get_stress_invariants(stress)
        assert pytest.approx(inv["I1"][0], abs=1e-6) == sigma
        assert pytest.approx(inv["I2"][0], abs=1e-6) == 0.0  # S11*S22 + ... - shear terms
        assert pytest.approx(inv["I3"][0], abs=1e-6) == 0.0

    def test_pure_shear(self):
        """纯剪切：I1=0"""
        tau = 50.0
        stress = np.array([[0, 0, 0, tau, 0, 0]])
        inv = get_stress_invariants(stress)
        assert pytest.approx(inv["I1"][0], abs=1e-6) == 0.0
        # I2 = -τ^2 (from -S12^2 term)
        assert pytest.approx(inv["I2"][0], abs=1e-6) == -tau ** 2

    def test_batch(self):
        """批量计算"""
        stress = np.random.rand(15, 6)
        inv = get_stress_invariants(stress)
        assert inv["I1"].shape == (15,)
        assert inv["I2"].shape == (15,)
        assert inv["I3"].shape == (15,)


class TestVonMisesFromUtils:
    """von_mises 函数测试（直接从 _utils 导入）"""

    def test_uniaxial_tension(self):
        """单轴拉伸：S11=100 → σ_vm = 100"""
        stress = np.array([[100.0, 0, 0, 0, 0, 0]])
        vm = von_mises(stress)
        assert pytest.approx(vm[0], abs=1e-6) == 100.0

    def test_hydrostatic_zero(self):
        """静水压：σ_vm = 0"""
        p = 50.0
        stress = np.array([[p, p, p, 0.0, 0.0, 0.0]])
        vm = von_mises(stress)
        assert pytest.approx(vm[0], abs=1e-6) == 0.0

    def test_pure_shear(self):
        """纯剪切：σ_vm = √3 · τ"""
        tau = 100.0
        stress = np.array([[0.0, 0.0, 0.0, tau, 0.0, 0.0]])
        vm = von_mises(stress)
        assert pytest.approx(vm[0], rel=1e-6) == tau * np.sqrt(3)

    def test_batch(self):
        """批量输入"""
        stress = np.random.rand(50, 6)
        vm = von_mises(stress)
        assert vm.shape == (50,)
        assert np.all(vm >= 0)


"""
CalculiX .dat 文件解析器

.dat 文件是 CalculiX 输出的文本格式结果文件，包含 NODE PRINT 和 EL PRINT 的结果。
本模块将其解析为 Python 数据结构。

.dat 文件结构速查：
  STEP X              — 载荷步开始
  INCREMENT X         — 增量步号
  displacements (...)  — 节点位移结果
  forces (...)        — 节点力/反力结果
  stresses (...)       — 单元应力结果
  strains (...)        — 单元应变结果
  contact ...         — 接触结果

参考 pygccx result_reader/dat_result.py 设计
"""
from __future__ import annotations

import re
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import numpy as np

from cae.enums import DatResultEntity


# 结果位置枚举
class ResultLocation(str):
    NODAL = "NODAL"      # 节点结果
    ELEMENT = "ELEMENT"  # 单元结果
    INT_PNT = "INT_PNT"  # 积分点结果


# 实体类型到结果位置的映射
ENTITY_2_LOCATION_MAP: dict[DatResultEntity, str] = {
    # 节点打印实体
    DatResultEntity.U: ResultLocation.NODAL,
    DatResultEntity.RF: ResultLocation.NODAL,
    # 单元打印实体
    DatResultEntity.S: ResultLocation.INT_PNT,
    DatResultEntity.E: ResultLocation.INT_PNT,
    DatResultEntity.ME: ResultLocation.INT_PNT,
    DatResultEntity.PEEQ: ResultLocation.INT_PNT,
    DatResultEntity.EVOL: ResultLocation.ELEMENT,
    DatResultEntity.COORD: ResultLocation.INT_PNT,
    DatResultEntity.ENER: ResultLocation.INT_PNT,
    DatResultEntity.ELKE: ResultLocation.ELEMENT,
    DatResultEntity.ELSE: ResultLocation.ELEMENT,
    DatResultEntity.EMAS: ResultLocation.ELEMENT,
    # 接触打印实体
    DatResultEntity.CELS: ResultLocation.NODAL,
    DatResultEntity.CSTR: ResultLocation.NODAL,
    DatResultEntity.CDIS: ResultLocation.NODAL,
}


# 分析类型
class DatAnalysisType:
    STATIC = "STATIC"
    FREQUENCY = "FREQUENCY"
    BUCKLE = "BUCKLE"


@dataclass(frozen=True)
class DatResultSet:
    """
    单个结果集。

    Attributes:
        entity: 结果实体类型（U, RF, S, E 等）
        no_components: 每个值的分量数
        step_time: 时间值
        step_no: 载荷步号
        step_inc_no: 增量步号（静态为增量号，模态为模态号，屈曲为屈曲因子号）
        analysis_type: 分析类型
        set_name: 节点集或单元集名称
        component_names: 分量名称元组
        values: 字典，key=节点ID/单元ID，value=分量值数组
        entity_location: 结果位置（NODAL/ELEMENT/INT_PNT）
    """
    entity: DatResultEntity
    no_components: int
    step_time: float
    step_no: int
    step_inc_no: int
    analysis_type: str
    set_name: str
    component_names: tuple[str, ...]
    values: dict[int, np.ndarray] = field(repr=False)
    entity_location: str = ResultLocation.NODAL

    def get_values_by_ids(self, ids: list[int]) -> np.ndarray:
        """
        按节点/单元 ID 列表获取结果值。

        Args:
            ids: 节点或单元 ID 列表

        Returns:
            如果 entity_location 是 NODAL 或 ELEMENT，返回 (N, M) 数组，N=ID数量，M=分量数
            如果 entity_location 是 INT_PNT，返回 (N, P, M) 数组，N=ID数量，P=积分点数量，M=分量数
        """
        return np.array([self.values[id] for id in ids])


@dataclass
class DatResult:
    """
    .dat 文件解析结果容器。

    提供 `get_result_sets_by()` 方法按条件过滤结果集。

    Usage:
        result = parse_dat("output.dat")
        # 获取所有位移结果
        u_sets = result.get_result_sets_by(entity=DatResultEntity.U)
        # 获取 step=1 的位移结果
        u_step1 = result.get_result_sets_by(entity=DatResultEntity.U, step_no=1)
        # 获取时间最接近 0.5 的结果
        u_near05 = result.get_result_sets_by(entity=DatResultEntity.U, step_time=0.5)
    """
    result_sets: list[DatResultSet] = field(default_factory=list)

    def get_result_sets_by(
        self,
        *,
        entity: Optional[DatResultEntity] = None,
        step_no: Optional[int] = None,
        step_inc_no: Optional[int] = None,
        step_time: Optional[float] = None,
        analysis_type: Optional[str] = None,
        set_name: Optional[str] = None,
    ) -> list[DatResultSet]:
        """
        按条件过滤结果集。

        所有参数都是可选的。step_time 会在最后应用，返回时间最接近的结果集。

        Args:
            entity: 结果实体类型（如 DatResultEntity.U, DatResultEntity.S）
            step_no: 载荷步号
            step_inc_no: 增量步号
            step_time: 目标时间（返回时间最接近的结果）
            analysis_type: 分析类型（STATIC, FREQUENCY, BUCKLE）
            set_name: 节点集或单元集名称

        Returns:
            符合条件的 DatResultSet 列表
        """
        rs = self.result_sets

        if entity is not None:
            rs = [r for r in rs if r.entity == entity]

        if step_no is not None:
            rs = [r for r in rs if r.step_no == step_no]

        if step_inc_no is not None:
            rs = [r for r in rs if r.step_inc_no == step_inc_no]

        if analysis_type is not None:
            rs = [r for r in rs if r.analysis_type.upper() == analysis_type.upper()]

        if set_name is not None:
            rs = [r for r in rs if r.set_name == set_name]

        if step_time is not None:
            available_times = {r.step_time for r in self.result_sets}
            nearest_time = min(available_times, key=lambda x: abs(x - step_time))
            rs = [r for r in rs if r.step_time == nearest_time]

        return rs

    def get_available_times(self) -> list[float]:
        """返回所有去重排序后的时间列表。"""
        return sorted({r.step_time for r in self.result_sets})

    def get_result_entities(self) -> list[DatResultEntity]:
        """返回所有结果实体类型的去重列表。"""
        return list({r.entity for r in self.result_sets})

    def get_set_names(self) -> list[str]:
        """返回所有节点集/单元集名称的去重列表。"""
        return list({r.set_name for r in self.result_sets if r.set_name})


def _is_numeric(s: str) -> bool:
    """检查字符串是否为数值。"""
    try:
        float(s)
        return True
    except ValueError:
        return False


def _parse_data_line(line: list[str], entity_loc: str) -> tuple[int, np.ndarray]:
    """
    解析数据行。

    Args:
        line: 分割后的行数据
        entity_loc: 结果位置

    Returns:
        (id, values_array)
    """
    id_ = int(line[0])  # 如果不是整数会抛出异常

    # 过滤掉非数值元素（如行尾的 'L'）
    numeric_values = [s for s in line[1:] if _is_numeric(s)]

    if entity_loc == ResultLocation.INT_PNT:
        # 积分点结果：id, int_pnt_id, value1, value2, ...
        # 实际格式可能是: elem_id int_pnt_id values...
        return id_, np.array([float(s) for s in numeric_values[1:]], dtype=float)

    return id_, np.array([float(s) for s in numeric_values], dtype=float)


def _parse_header_line(line: list[str]) -> tuple[str, str, float, tuple[str, ...]]:
    """
    解析结果头行。

    格式示例：
        displacements (vx,vy,vz) for set SET1 and time  0.1000000E+01
        stresses (elem, integ.pnt.,sxx,syy,szz,sxy,sxz,syz) for set SET2 and time  0.1000000E+01

    Returns:
        (entity_name, set_name, step_time, component_names)
    """
    line_text = ' '.join(line)

    # 跳过 total 行
    if line[0].lower() == 'total':
        raise ValueError("Skipping total line")

    # 解析时间值（最后一个数值）
    step_time = float(line[-1])

    # 解析 set_name
    set_name = ''
    try:
        idx_and = line.index('and')
        try:
            idx_set = line.index('set')
            set_name = ' '.join(line[idx_set + 1:idx_and])
        except ValueError:
            pass
    except ValueError:
        pass

    # 解析实体名称和分量
    # 实体名在 '(' 之前或 'for' 之前
    entity_name = ''
    component_names: tuple[str, ...] = ()

    # 提取分量名
    comp_match = re.findall(r'\((.*?)\)', line_text)
    if comp_match:
        # 处理分量名中的逗号分隔
        all_components = []
        for c in comp_match:
            # 有些格式是 "elem, integ.pnt.,sxx,syy,szz,sxy,sxz,syz"
            parts = c.replace(',', ' ').split()
            all_components.extend([p.strip() for p in parts if p.strip()])

        component_names = tuple(all_components)

    # 提取实体名
    for i, col in enumerate(line):
        if col.startswith('(') or col == 'for':
            entity_name = ' '.join(line[:i])
            break

    # 标准化实体名
    entity_name_lower = entity_name.lower()
    if 'displacement' in entity_name_lower:
        entity_name = 'U'
    elif 'force' in entity_name_lower or 'reaction' in entity_name_lower:
        entity_name = 'RF'
    elif 'stress' in entity_name_lower:
        entity_name = 'S'
    elif 'strain' in entity_name_lower and 'mechanical' not in entity_name_lower:
        entity_name = 'E'
    elif 'mechanical strain' in entity_name_lower:
        entity_name = 'ME'
    elif 'equivalent plastic' in entity_name_lower or 'peeq' in entity_name_lower:
        entity_name = 'PEEQ'
    elif 'volume' in entity_name_lower:
        entity_name = 'EVOL'
    elif 'coordinate' in entity_name_lower:
        entity_name = 'COORD'
    elif 'internal energy density' in entity_name_lower:
        entity_name = 'ENER'
    elif 'kinetic energy' in entity_name_lower:
        entity_name = 'ELKE'
    elif 'internal energy' in entity_name_lower:
        entity_name = 'ELSE'
    elif 'mass' in entity_name_lower:
        entity_name = 'EMAS'
    elif 'contact energy' in entity_name_lower:
        entity_name = 'CELS'
    elif 'contact stress' in entity_name_lower:
        entity_name = 'CSTR'
    elif 'contact displacement' in entity_name_lower or 'relative contact' in entity_name_lower:
        entity_name = 'CDIS'

    return entity_name, set_name, step_time, component_names


def _value_dict_to_array(
    value_dict: dict[int, list[np.ndarray]], entity_loc: str
) -> dict[int, np.ndarray]:
    """将值字典转换为 numpy 数组。"""
    if entity_loc == ResultLocation.INT_PNT:
        # 积分点：每个 ID 对应多个积分点
        return {id_: np.array(lst) for id_, lst in value_dict.items()}
    # 节点/单元：每个 ID 对应一个值数组
    return {id_: lst[0] for id_, lst in value_dict.items()}


def _get_no_components(value_dict: dict[int, np.ndarray]) -> int:
    """获取分量数。"""
    arr = next(iter(value_dict.values()))
    if len(arr.shape) == 1:
        return arr.shape[0]
    # 积分点情况：返回每行分量数
    return arr.shape[1] if arr.ndim > 1 else 1


def _detect_entity_type(entity_name: str) -> DatResultEntity:
    """从实体名称检测实体类型枚举。"""
    name_map = {
        'U': DatResultEntity.U,
        'RF': DatResultEntity.RF,
        'S': DatResultEntity.S,
        'E': DatResultEntity.E,
        'ME': DatResultEntity.ME,
        'PEEQ': DatResultEntity.PEEQ,
        'EVOL': DatResultEntity.EVOL,
        'COORD': DatResultEntity.COORD,
        'ENER': DatResultEntity.ENER,
        'ELKE': DatResultEntity.ELKE,
        'ELSE': DatResultEntity.ELSE,
        'EMAS': DatResultEntity.EMAS,
        'CELS': DatResultEntity.CELS,
        'CSTR': DatResultEntity.CSTR,
        'CDIS': DatResultEntity.CDIS,
    }
    return name_map.get(entity_name.upper(), DatResultEntity.U)


class DatReader:
    """
    .dat 文件状态机解析器。

    使用方法：
        reader = DatReader()
        result = reader.parse_file("output.dat")
    """

    class State:
        NONE = "NONE"
        RESULT_SET_OPEN = "RESULT_SET_OPEN"

    def __init__(self):
        self.state = self.State.NONE
        self.result_sets: list[DatResultSet] = []

        # 当前结果集的状态
        self.entity_name = ''
        self.set_name = ''
        self.step_time = 0.0
        self.step_no = 0
        self.step_inc = 0
        self.analysis_type = DatAnalysisType.STATIC
        self.component_names: tuple[str, ...] = ()
        self.value_dict: dict[int, list[np.ndarray]] = defaultdict(list)

    def parse_file(self, filepath: str | Path) -> DatResult:
        """
        解析 .dat 文件。

        Args:
            filepath: .dat 文件路径

        Returns:
            DatResult 对象

        Raises:
            FileNotFoundError: 文件不存在
            ValueError: 文件格式不支持（ccx < 2.22）
        """
        with open(filepath, encoding='utf-8', errors='replace') as f:
            lines = [line.strip() for line in f if line.strip()]

        return self._parse_lines(lines)

    def _parse_lines(self, lines: list[str]) -> DatResult:
        """解析行列表。"""
        self.result_sets = []
        self.step_no = 0
        self.step_inc = 0
        self.analysis_type = DatAnalysisType.STATIC

        has_step_word = False  # ccx >= 2.22 的标志

        i = 0
        while i < len(lines):
            line = lines[i]
            parts = line.split()

            if not parts:
                i += 1
                continue

            # 状态机
            if self.state == self.State.NONE:
                first_char = parts[0][0].upper() if parts[0] else ''

                # 检测 STEP 开始（可能有空格，如 "S T E P"）
                line_upper = line.upper().replace(' ', '')
                if first_char == 'S' and line_upper.startswith('STEP'):
                    has_step_word = True
                    self.step_no = int(parts[-1])
                    self.step_inc = 0
                    self.analysis_type = DatAnalysisType.STATIC

                # 检测 EIGENVALUENUMBER（模态/屈曲步）
                elif first_char == 'E' and line_upper.startswith('EIGENVALUENUMBER'):
                    self.step_inc = int(parts[-1])

                # 检测 INCREMENT（静态增量步）
                elif first_char == 'I' and line_upper.startswith('INCREMENT'):
                    self.step_inc = int(parts[-1])

                # 检测频率信息
                elif first_char == 'E' and line_upper.startswith('EIGENVALUEOUTPUT'):
                    self.analysis_type = DatAnalysisType.FREQUENCY

                # 检测屈曲信息
                elif first_char == 'B' and line_upper.startswith('BUCKLINGFACTOROUTPUT'):
                    self.analysis_type = DatAnalysisType.BUCKLE

                # 检测结果集开始
                elif self._is_result_header(line):
                    if not has_step_word:
                        raise ValueError(
                            "此 .dat 文件由 ccx < 2.22 生成，不支持解析。"
                            "请使用 ccx >= 2.22。"
                        )
                    try:
                        self._start_result_set(parts)
                        self.state = self.State.RESULT_SET_OPEN
                    except Exception:
                        pass

            elif self.state == self.State.RESULT_SET_OPEN:
                # 尝试解析数据行
                try:
                    entity_loc = ENTITY_2_LOCATION_MAP.get(
                        _detect_entity_type(self.entity_name),
                        ResultLocation.NODAL
                    )
                    id_, values = _parse_data_line(parts, entity_loc)
                    self.value_dict[id_].append(values)
                except (ValueError, IndexError):
                    # 不是数据行，结束当前结果集
                    self._finish_result_set()
                    self.state = self.State.NONE
                    continue

            i += 1

        # 处理最后的结果集
        if self.state == self.State.RESULT_SET_OPEN:
            self._finish_result_set()

        return DatResult(result_sets=self.result_sets)

    def _is_result_header(self, line: str) -> bool:
        """检查行是否为结果头行。"""
        # 常见结果类型关键词
        keywords = [
            'displacement', 'force', 'stress', 'strain', 'coordinate',
            'volume', 'energy', 'mass', 'contact', 'plastic'
        ]
        # 移除空格后转小写
        line_normalized = line.lower().replace(' ', '')
        return any(kw in line_normalized for kw in keywords)

    def _start_result_set(self, line: list[str]) -> None:
        """开始新的结果集。"""
        self.entity_name, self.set_name, self.step_time, self.component_names = \
            _parse_header_line(line)

        self.value_dict = defaultdict(list)

    def _finish_result_set(self) -> None:
        """完成结果集，添加到列表。"""
        if not self.value_dict:
            return

        entity_type = _detect_entity_type(self.entity_name)
        entity_loc = ENTITY_2_LOCATION_MAP.get(entity_type, ResultLocation.NODAL)

        values_arr = _value_dict_to_array(self.value_dict, entity_loc)
        no_comp = _get_no_components(values_arr)

        result_set = DatResultSet(
            entity=entity_type,
            no_components=no_comp,
            step_time=self.step_time,
            step_no=self.step_no,
            step_inc_no=self.step_inc,
            analysis_type=self.analysis_type,
            set_name=self.set_name,
            component_names=self.component_names[-no_comp:] if self.component_names else (),
            values=values_arr,
            entity_location=entity_loc,
        )
        self.result_sets.append(result_set)


def parse_dat(filepath: str | Path) -> DatResult:
    """
    解析 .dat 文件。

    Args:
        filepath: .dat 文件路径

    Returns:
        DatResult 对象

    Usage:
        result = parse_dat("output.dat")

        # 获取所有位移
        u_sets = result.get_result_sets_by(entity=DatResultEntity.U)

        # 获取 step=1 的应力
        s_sets = result.get_result_sets_by(entity=DatResultEntity.S, step_no=1)

        # 获取时间最接近 0.5 的结果
        sets_near_05 = result.get_result_sets_by(step_time=0.5)
    """
    reader = DatReader()
    return reader.parse_file(filepath)
